@@ -37,6 +37,8 @@
 #include <math.h>
 #include <string.h>
 
+#include <gio/gio.h>
+
 #include "pango-types.h"
 #include "pango-font-private.h"
 #include "pango-fontmap.h"
@@ -1797,7 +1799,7 @@ pango_font_get_glyph_extents  (PangoFont      *font,
 	}
       if (logical_rect)
 	{
-	  logical_rect->x = logical_rect->y = 0;
+	  logical_rect->x = 0;
 	  logical_rect->y = - PANGO_UNKNOWN_GLYPH_HEIGHT * PANGO_SCALE;
 	  logical_rect->height = PANGO_UNKNOWN_GLYPH_HEIGHT * PANGO_SCALE;
 	  logical_rect->width = PANGO_UNKNOWN_GLYPH_WIDTH * PANGO_SCALE;
@@ -1880,6 +1882,24 @@ pango_font_get_font_map (PangoFont *font)
     return NULL;
 }
 
+/**
+ * pango_font_get_face:
+ * @font: a #PangoFont
+ *
+ * Gets the #PangoFontFace to which @font belongs.
+ *
+ * Returns: (transfer none): the #PangoFontFace
+ *
+ * Since: 1.46
+ */
+PangoFontFace *
+pango_font_get_face (PangoFont *font)
+{
+  PangoFontMap *map = pango_font_get_font_map (font);
+
+  return PANGO_FONT_MAP_GET_CLASS (map)->get_face (map,font);
+}
+
 /**
  * pango_font_get_hb_font: (skip)
  * @font: a #PangoFont
@@ -2158,11 +2178,62 @@ pango_font_metrics_get_strikethrough_thickness (PangoFontMetrics *metrics)
  * PangoFontFamily
  */
 
-G_DEFINE_ABSTRACT_TYPE (PangoFontFamily, pango_font_family, G_TYPE_OBJECT)
+static GType
+pango_font_family_get_item_type (GListModel *list)
+{
+  return PANGO_TYPE_FONT_FACE;
+}
+
+static guint
+pango_font_family_get_n_items (GListModel *list)
+{
+  PangoFontFamily *family = PANGO_FONT_FAMILY (list);
+  int n_faces;
+
+  pango_font_family_list_faces (family, NULL, &n_faces);
+
+  return (guint)n_faces;
+}
+
+static gpointer
+pango_font_family_get_item (GListModel *list,
+                            guint       position)
+{
+  PangoFontFamily *family = PANGO_FONT_FAMILY (list);
+  PangoFontFace **faces;
+  int n_faces;
+  PangoFontFace *face;
+
+  pango_font_family_list_faces (family, &faces, &n_faces);
+
+  if (position < n_faces)
+    face = g_object_ref (faces[position]);
+  else
+    face = NULL;
+
+  g_free (faces);
+
+  return face;
+}
+
+static void
+pango_font_family_list_model_init (GListModelInterface *iface)
+{
+  iface->get_item_type = pango_font_family_get_item_type;
+  iface->get_n_items = pango_font_family_get_n_items;
+  iface->get_item = pango_font_family_get_item;
+}
+
+G_DEFINE_ABSTRACT_TYPE_WITH_CODE (PangoFontFamily, pango_font_family, G_TYPE_OBJECT,
+                                  G_IMPLEMENT_INTERFACE (G_TYPE_LIST_MODEL, pango_font_family_list_model_init))
+
+static PangoFontFace *pango_font_family_real_get_face (PangoFontFamily *family,
+                                                       const char      *name);
 
 static void
 pango_font_family_class_init (PangoFontFamilyClass *class G_GNUC_UNUSED)
 {
+  class->get_face = pango_font_family_real_get_face;
 }
 
 static void
@@ -2212,6 +2283,62 @@ pango_font_family_list_faces (PangoFontFamily  *family,
   PANGO_FONT_FAMILY_GET_CLASS (family)->list_faces (family, faces, n_faces);
 }
 
+static PangoFontFace *
+pango_font_family_real_get_face (PangoFontFamily *family,
+                                 const char      *name)
+{
+  PangoFontFace **faces;
+  int n_faces;
+  PangoFontFace *face;
+  int i;
+
+  pango_font_family_list_faces (family, &faces, &n_faces);
+
+  face = NULL;
+  if (name == NULL)
+    {
+      face = faces[0];
+    }
+  else
+    {
+      for (i = 0; i < n_faces; i++)
+        {
+          if (strcmp (name, pango_font_face_get_face_name (faces[i])) == 0)
+            {
+              face = faces[i];
+              break;
+            }
+        }
+    }
+
+  g_free (faces);
+
+  return face;
+}
+
+/**
+ * pango_font_family_get_face:
+ * @family: a #PangoFontFamily
+ * @name: (nullable): the name of a face. If the name is %NULL,
+ *     the family's default face (fontconfig calls it "Regular")
+ *     will be returned.
+ *
+ * Gets the #PangoFontFace of @family with the given name.
+ *
+ * Returns: (transfer none) (nullable): the #PangoFontFace,
+ *     or %NULL if no face with the given name exists.
+ *
+ * Since: 1.46
+ */
+PangoFontFace *
+pango_font_family_get_face (PangoFontFamily *family,
+                            const char      *name)
+{
+  g_return_val_if_fail (PANGO_IS_FONT_FAMILY (family), NULL);
+
+  return PANGO_FONT_FAMILY_GET_CLASS (family)->get_face (family, name);
+}
+
 /**
  * pango_font_family_is_monospace:
  * @family: a #PangoFontFamily
@@ -2381,6 +2508,25 @@ pango_font_face_list_sizes (PangoFontFace  *face,
     }
 }
 
+/**
+ * pango_font_face_get_family:
+ * @face: a #PangoFontFace
+ *
+ * Gets the #PangoFontFamily that @face
+ * belongs to.
+ *
+ * Returns: (transfer none): the #PangoFontFamily
+ *
+ * Since: 1.46
+ */
+PangoFontFamily *
+pango_font_face_get_family (PangoFontFace *face)
+{
+  g_return_val_if_fail (PANGO_IS_FONT_FACE (face), NULL);
+
+  return PANGO_FONT_FACE_GET_CLASS (face)->get_family (face);
+}
+
 /**
  * pango_font_has_char:
  * @font: a #PangoFont
