@@ -28,6 +28,8 @@
  * Unicode characters and converts it into glyphs.
  * The functions described in this section accomplish
  * various steps of this process.
+ *
+ * ![](pipeline.png)
  */
 #include "config.h"
 #include <string.h>
@@ -36,8 +38,22 @@
 #include "pango-context.h"
 #include "pango-impl-utils.h"
 
-#include "pango-engine-private.h"
+#include "pango-font-private.h"
+#include "pango-fontset-private.h"
+#include "pango-fontmap-private.h"
 #include "pango-script-private.h"
+#include "pango-emoji-private.h"
+
+/**
+ * SECTION:context
+ * @title:Contexts
+ * @short_description: Global context object
+ *
+ * The #PangoContext structure stores global information
+ * influencing Pango's operation, such as the fontmap used
+ * to look up fonts, and default values such as the default
+ * language, default gravity, or default font.
+ */
 
 /**
  * PangoContext:
@@ -63,6 +79,8 @@ struct _PangoContext
   PangoMatrix *matrix;
 
   PangoFontMap *font_map;
+
+  gboolean round_glyph_positions;
 };
 
 struct _PangoContextClass
@@ -87,6 +105,7 @@ pango_context_init (PangoContext *context)
   context->set_language = NULL;
   context->language = pango_language_get_default ();
   context->font_map = NULL;
+  context->round_glyph_positions = TRUE;
 
   context->font_desc = pango_font_description_new ();
   pango_font_description_set_family_static (context->font_desc, "serif");
@@ -600,7 +619,7 @@ advance_attr_iterator_to (PangoAttrIterator *iterator,
 }
 
 /***************************************************************************
- * We cache the results of character,fontset => font,shaper in a hash table
+ * We cache the results of character,fontset => font in a hash table
  ***************************************************************************/
 
 typedef struct {
@@ -633,7 +652,7 @@ get_font_cache (PangoFontset *fontset)
 
   static GQuark cache_quark = 0; /* MT-safe */
   if (G_UNLIKELY (!cache_quark))
-    cache_quark = g_quark_from_static_string ("pango-shaper-font-cache");
+    cache_quark = g_quark_from_static_string ("pango-font-cache");
 
 retry:
   cache = g_object_get_qdata (G_OBJECT (fontset), cache_quark);
@@ -691,7 +710,8 @@ typedef enum {
   LANG_CHANGED         = 1 << 2,
   FONT_CHANGED         = 1 << 3,
   DERIVED_LANG_CHANGED = 1 << 4,
-  WIDTH_CHANGED        = 1 << 5
+  WIDTH_CHANGED        = 1 << 5,
+  EMOJI_CHANGED        = 1 << 6,
 } ChangedFlags;
 
 
@@ -704,7 +724,7 @@ struct _PangoWidthIter
 	const gchar *text_end;
 	const gchar *start;
 	const gchar *end;
-	gboolean wide;
+	gboolean     upright;
 };
 
 typedef struct _ItemizeState ItemizeState;
@@ -738,6 +758,7 @@ struct _ItemizeState
   gboolean free_attr_iter;
   const char *attr_end;
   PangoFontDescription *font_desc;
+  PangoFontDescription *emoji_font_desc;
   PangoLanguage *lang;
   GSList *extra_attrs;
   gboolean copy_extra_attrs;
@@ -749,9 +770,9 @@ struct _ItemizeState
   PangoScript script;
 
   PangoWidthIter width_iter;
+  PangoEmojiIter emoji_iter;
 
   PangoLanguage *derived_lang;
-  PangoEngineLang *lang_engine;
 
   PangoFontset *current_fonts;
   FontCache *cache;
@@ -799,6 +820,12 @@ update_attr_iterator (ItemizeState *state)
   else
     state->attr_end = state->end;
 
+  if (state->emoji_font_desc)
+    {
+      pango_font_description_free (state->emoji_font_desc);
+      state->emoji_font_desc = NULL;
+    }
+
   old_lang = state->lang;
   if (state->font_desc)
     pango_font_description_free (state->font_desc);
@@ -839,44 +866,106 @@ update_end (ItemizeState *state)
     state->run_end = state->script_end;
   if (state->width_iter.end < state->run_end)
     state->run_end = state->width_iter.end;
+  if (state->emoji_iter.end < state->run_end)
+    state->run_end = state->emoji_iter.end;
 }
 
-/* g_unichar_iswide() uses EastAsianWidth, which is broken.
- * We should switch to using VerticalTextLayout:
- * http://www.unicode.org/reports/tr50/#Data50
- *
- * In the mean time, fixup Hangul jamo to be all wide so we
- * don't break run in the middle.  The EastAsianWidth has
- * 'W' for L-jamo, and 'N' for T and V jamo!
- *
- * https://bugzilla.gnome.org/show_bug.cgi?id=705727
- */
 static gboolean
-width_iter_iswide (gunichar ch)
+width_iter_is_upright (gunichar ch)
 {
-  if ((0x1100u <= ch && ch <= 0x11FFu) ||
-      (0xA960u <= ch && ch <= 0xA97Cu) ||
-      (0xD7B0u <= ch && ch <= 0xD7FBu))
-    return TRUE;
+  /* https://www.unicode.org/Public/11.0.0/ucd/VerticalOrientation.txt
+   * VO=U or Tu table generated by tools/gen-vertical-orientation-U-table.py.
+   *
+   * FIXME: In the future, If GLib supports VerticalOrientation, please use it.
+   */
+  static const gunichar upright[][2] = {
+    {0x00A7, 0x00A7}, {0x00A9, 0x00A9}, {0x00AE, 0x00AE}, {0x00B1, 0x00B1},
+    {0x00BC, 0x00BE}, {0x00D7, 0x00D7}, {0x00F7, 0x00F7}, {0x02EA, 0x02EB},
+    {0x1100, 0x11FF}, {0x1401, 0x167F}, {0x18B0, 0x18FF}, {0x2016, 0x2016},
+    {0x2020, 0x2021}, {0x2030, 0x2031}, {0x203B, 0x203C}, {0x2042, 0x2042},
+    {0x2047, 0x2049}, {0x2051, 0x2051}, {0x2065, 0x2065}, {0x20DD, 0x20E0},
+    {0x20E2, 0x20E4}, {0x2100, 0x2101}, {0x2103, 0x2109}, {0x210F, 0x210F},
+    {0x2113, 0x2114}, {0x2116, 0x2117}, {0x211E, 0x2123}, {0x2125, 0x2125},
+    {0x2127, 0x2127}, {0x2129, 0x2129}, {0x212E, 0x212E}, {0x2135, 0x213F},
+    {0x2145, 0x214A}, {0x214C, 0x214D}, {0x214F, 0x2189}, {0x218C, 0x218F},
+    {0x221E, 0x221E}, {0x2234, 0x2235}, {0x2300, 0x2307}, {0x230C, 0x231F},
+    {0x2324, 0x2328}, {0x232B, 0x232B}, {0x237D, 0x239A}, {0x23BE, 0x23CD},
+    {0x23CF, 0x23CF}, {0x23D1, 0x23DB}, {0x23E2, 0x2422}, {0x2424, 0x24FF},
+    {0x25A0, 0x2619}, {0x2620, 0x2767}, {0x2776, 0x2793}, {0x2B12, 0x2B2F},
+    {0x2B50, 0x2B59}, {0x2BB8, 0x2BD1}, {0x2BD3, 0x2BEB}, {0x2BF0, 0x2BFF},
+    {0x2E80, 0x3007}, {0x3012, 0x3013}, {0x3020, 0x302F}, {0x3031, 0x309F},
+    {0x30A1, 0x30FB}, {0x30FD, 0xA4CF}, {0xA960, 0xA97F}, {0xAC00, 0xD7FF},
+    {0xE000, 0xFAFF}, {0xFE10, 0xFE1F}, {0xFE30, 0xFE48}, {0xFE50, 0xFE57},
+    {0xFE5F, 0xFE62}, {0xFE67, 0xFE6F}, {0xFF01, 0xFF07}, {0xFF0A, 0xFF0C},
+    {0xFF0E, 0xFF19}, {0xFF1F, 0xFF3A}, {0xFF3C, 0xFF3C}, {0xFF3E, 0xFF3E},
+    {0xFF40, 0xFF5A}, {0xFFE0, 0xFFE2}, {0xFFE4, 0xFFE7}, {0xFFF0, 0xFFF8},
+    {0xFFFC, 0xFFFD}, {0x10980, 0x1099F}, {0x11580, 0x115FF}, {0x11A00, 0x11AAF},
+    {0x13000, 0x1342F}, {0x14400, 0x1467F}, {0x16FE0, 0x18AFF}, {0x1B000, 0x1B12F},
+    {0x1B170, 0x1B2FF}, {0x1D000, 0x1D1FF}, {0x1D2E0, 0x1D37F}, {0x1D800, 0x1DAAF},
+    {0x1F000, 0x1F7FF}, {0x1F900, 0x1FA6F}, {0x20000, 0x2FFFD}, {0x30000, 0x3FFFD},
+    {0xF0000, 0xFFFFD}, {0x100000, 0x10FFFD}
+  };
+  static const int max = sizeof(upright) / sizeof(upright[0]);
+  int st = 0;
+  int ed = max;
 
-  return g_unichar_iswide (ch);
+  while (st <= ed)
+    {
+      int mid = (st + ed) / 2;
+      if (upright[mid][0] <= ch && ch <= upright[mid][1])
+        return TRUE;
+      else
+        if (upright[mid][0] <= ch)
+          st = mid + 1;
+        else
+          ed = mid - 1;
+    }
+
+  return FALSE;
 }
 
 static void
 width_iter_next(PangoWidthIter* iter)
 {
+  gboolean met_joiner = FALSE;
   iter->start = iter->end;
 
   if (iter->end < iter->text_end)
     {
       gunichar ch = g_utf8_get_char (iter->end);
-      iter->wide = width_iter_iswide (ch);
+      iter->upright = width_iter_is_upright (ch);
     }
 
   while (iter->end < iter->text_end)
     {
       gunichar ch = g_utf8_get_char (iter->end);
-      if (width_iter_iswide (ch) != iter->wide)
+
+      /* for zero width joiner */
+      if (ch == 0x200D)
+        {
+          iter->end = g_utf8_next_char (iter->end);
+          met_joiner = TRUE;
+          continue;
+        }
+
+      /* ignore the upright check if met joiner */
+      if (met_joiner)
+        {
+          iter->end = g_utf8_next_char (iter->end);
+          met_joiner = FALSE;
+          continue;
+        }
+
+      /* for variation selector, tag and emoji modifier. */
+      if (G_UNLIKELY(ch == 0xFE0EU || ch == 0xFE0FU
+                    || (ch >= 0xE0020 && ch <= 0xE007F)
+                    || (ch >= 0x1F3FB && ch <= 0x1F3FF)))
+        {
+          iter->end = g_utf8_next_char (iter->end);
+          continue;
+        }
+
+      if (width_iter_is_upright (ch) != iter->upright)
         break;
       iter->end = g_utf8_next_char (iter->end);
     }
@@ -892,6 +981,11 @@ width_iter_init (PangoWidthIter* iter, const char* text, int length)
   width_iter_next (iter);
 }
 
+static void
+width_iter_fini (PangoWidthIter* iter)
+{
+}
+
 static void
 itemize_state_init (ItemizeState      *state,
 		    PangoContext      *context,
@@ -912,6 +1006,8 @@ itemize_state_init (ItemizeState      *state,
   state->item = NULL;
 
   state->run_start = text + start_index;
+  state->changed = EMBEDDING_CHANGED | SCRIPT_CHANGED | LANG_CHANGED |
+                   FONT_CHANGED | WIDTH_CHANGED | EMOJI_CHANGED;
 
   /* First, apply the bidirectional algorithm to break
    * the text into directional runs.
@@ -940,6 +1036,7 @@ itemize_state_init (ItemizeState      *state,
       state->free_attr_iter = FALSE;
     }
 
+  state->emoji_font_desc = NULL;
   if (state->attr_iter)
     {
       state->font_desc = NULL;
@@ -965,8 +1062,11 @@ itemize_state_init (ItemizeState      *state,
   pango_script_iter_get_range (&state->script_iter, NULL,
 			       &state->script_end, &state->script);
 
-  /* Initialize the width iterator */
   width_iter_init (&state->width_iter, text + start_index, length);
+  _pango_emoji_iter_init (&state->emoji_iter, text + start_index, length);
+
+  if (state->emoji_iter.is_emoji)
+    state->width_iter.end = MAX (state->width_iter.end, state->emoji_iter.end);
 
   update_end (state);
 
@@ -980,12 +1080,10 @@ itemize_state_init (ItemizeState      *state,
   state->gravity_hint = state->context->gravity_hint;
   state->resolved_gravity = PANGO_GRAVITY_AUTO;
   state->derived_lang = NULL;
-  state->lang_engine = NULL;
   state->current_fonts = NULL;
   state->cache = NULL;
   state->base_font = NULL;
 
-  state->changed = EMBEDDING_CHANGED | SCRIPT_CHANGED | LANG_CHANGED | FONT_CHANGED | WIDTH_CHANGED;
 }
 
 static gboolean
@@ -1016,6 +1114,14 @@ itemize_state_next (ItemizeState *state)
 				   &state->script_end, &state->script);
       state->changed |= SCRIPT_CHANGED;
     }
+  if (state->run_end == state->emoji_iter.end)
+    {
+      _pango_emoji_iter_next (&state->emoji_iter);
+      state->changed |= EMOJI_CHANGED;
+
+      if (state->emoji_iter.is_emoji)
+        state->width_iter.end = MAX (state->width_iter.end, state->emoji_iter.end);
+    }
   if (state->run_end == state->width_iter.end)
     {
       width_iter_next (&state->width_iter);
@@ -1040,47 +1146,39 @@ copy_attr_slist (GSList *attr_slist)
 }
 
 static void
-itemize_state_fill_shaper (ItemizeState     *state,
-			   PangoEngineShape *shape_engine,
-			   PangoFont        *font)
+itemize_state_fill_font (ItemizeState *state,
+			 PangoFont    *font)
 {
   GList *l;
 
   for (l = state->result; l; l = l->next)
     {
       PangoItem *item = l->data;
-      if (item->analysis.shape_engine)
-	break;
+      if (item->analysis.font)
+        break;
       if (font)
 	item->analysis.font = g_object_ref (font);
-      else
-	item->analysis.font = NULL;
-      item->analysis.shape_engine = shape_engine;
     }
 }
 
 static void
-itemize_state_add_character (ItemizeState     *state,
-			     PangoEngineShape *shape_engine,
-			     PangoFont        *font,
-			     gboolean          force_break,
-			     const char       *pos)
+itemize_state_add_character (ItemizeState *state,
+			     PangoFont    *font,
+			     gboolean      force_break,
+			     const char   *pos)
 {
   if (state->item)
     {
-      if (!state->item->analysis.shape_engine && shape_engine)
+      if (!state->item->analysis.font && font)
 	{
-	  itemize_state_fill_shaper (state, shape_engine, font);
+	  itemize_state_fill_font (state, font);
 	}
-      else if (state->item->analysis.shape_engine && !shape_engine)
+      else if (state->item->analysis.font && !font)
 	{
 	  font = state->item->analysis.font;
-	  shape_engine = state->item->analysis.shape_engine;
 	}
 
       if (!force_break &&
-	  state->item->analysis.lang_engine == state->lang_engine &&
-	  state->item->analysis.shape_engine == shape_engine &&
 	  state->item->analysis.font == font)
 	{
 	  state->item->num_chars++;
@@ -1094,8 +1192,6 @@ itemize_state_add_character (ItemizeState     *state,
   state->item->offset = pos - state->text;
   state->item->length = 0;
   state->item->num_chars = 1;
-  state->item->analysis.shape_engine = shape_engine;
-  state->item->analysis.lang_engine = state->lang_engine;
 
   if (font)
     g_object_ref (font);
@@ -1165,15 +1261,11 @@ get_font_foreach (PangoFontset *fontset,
 		  gpointer      data)
 {
   GetFontInfo *info = data;
-  PangoEngineShape *engine;
-  PangoCoverageLevel level;
 
   if (G_UNLIKELY (!font))
     return FALSE;
 
-  engine = pango_font_find_shaper (font, info->lang, info->wc),
-  level = _pango_engine_shape_covers (engine, font, info->lang, info->wc);
-  if (level != PANGO_COVERAGE_NONE)
+  if (pango_font_has_char (font, info->wc))
     {
       info->font = font;
       return TRUE;
@@ -1198,36 +1290,17 @@ get_base_font (ItemizeState *state)
   return state->base_font;
 }
 
-static PangoScript
-get_script (ItemizeState      *state)
-{
-  /* Always use a basic shaper for vertical layout (ie, east/west gravity)
-   * as none of our script shapers support vertical shaping right now.
-   *
-   * XXX Should move the knowledge into the shaper interface.
-   */
-
-  if (PANGO_GRAVITY_IS_VERTICAL (state->resolved_gravity))
-    return PANGO_SCRIPT_COMMON;
-  else
-    return state->script;
-}
-
 static gboolean
-get_shaper_and_font (ItemizeState      *state,
-		     gunichar           wc,
-		     PangoEngineShape **shape_engine,
-		     PangoFont        **font)
+get_font (ItemizeState  *state,
+          gunichar       wc,
+          PangoFont    **font)
 {
   GetFontInfo info;
 
   /* We'd need a separate cache when fallback is disabled, but since lookup
    * with fallback disabled is faster anyways, we just skip caching */
   if (state->enable_fallback && font_cache_get (state->cache, wc, font))
-  {
-    *shape_engine = pango_font_find_shaper (*font, state->derived_lang, wc);
     return TRUE;
-  }
 
   info.lang = state->derived_lang;
   info.wc = wc;
@@ -1239,7 +1312,6 @@ get_shaper_and_font (ItemizeState      *state,
     get_font_foreach (NULL, get_base_font (state), &info);
 
   *font = info.font;
-  *shape_engine = pango_font_find_shaper (*font, state->derived_lang, wc);
 
   /* skip caching if fallback disabled (see above) */
   if (state->enable_fallback)
@@ -1284,8 +1356,6 @@ itemize_state_update_for_new_run (ItemizeState *state)
    * do it right now */
   if (state->changed & (FONT_CHANGED | SCRIPT_CHANGED | WIDTH_CHANGED))
     {
-      PangoGravity old_gravity = state->resolved_gravity;
-
       /* Font-desc gravity overrides everything */
       if (state->font_desc_gravity != PANGO_GRAVITY_AUTO)
 	{
@@ -1300,12 +1370,12 @@ itemize_state_update_for_new_run (ItemizeState *state)
 	    gravity = state->context->resolved_gravity;
 
 	  state->resolved_gravity = pango_gravity_get_for_script_and_width (state->script,
-									    state->width_iter.wide,
+									    state->width_iter.upright,
 									    gravity,
 									    gravity_hint);
 	}
 
-      if (old_gravity != state->resolved_gravity)
+      if (state->font_desc_gravity != state->resolved_gravity)
 	{
 	  pango_font_description_set_gravity (state->font_desc, state->resolved_gravity);
 	  state->changed |= FONT_CHANGED;
@@ -1320,9 +1390,9 @@ itemize_state_update_for_new_run (ItemizeState *state)
 	state->changed |= DERIVED_LANG_CHANGED;
     }
 
-  if ((state->changed & DERIVED_LANG_CHANGED) || !state->lang_engine)
+  if (state->changed & (EMOJI_CHANGED))
     {
-      state->lang_engine = _pango_get_language_engine ();
+      state->changed |= FONT_CHANGED;
     }
 
   if (state->changed & (FONT_CHANGED | DERIVED_LANG_CHANGED) &&
@@ -1335,9 +1405,15 @@ itemize_state_update_for_new_run (ItemizeState *state)
 
   if (!state->current_fonts)
     {
+      gboolean is_emoji = state->emoji_iter.is_emoji;
+      if (is_emoji && !state->emoji_font_desc)
+      {
+        state->emoji_font_desc = pango_font_description_copy_static (state->font_desc);
+        pango_font_description_set_family_static (state->emoji_font_desc, "emoji");
+      }
       state->current_fonts = pango_font_map_load_fontset (state->context->font_map,
 							  state->context,
-							  state->font_desc,
+							  is_emoji ? state->emoji_font_desc : state->font_desc,
 							  state->derived_lang);
       state->cache = get_font_cache (state->current_fonts);
     }
@@ -1349,21 +1425,6 @@ itemize_state_update_for_new_run (ItemizeState *state)
     }
 }
 
-static const char *
-string_from_script (PangoScript script)
-{
-  static GEnumClass *class = NULL; /* MT-safe */
-  GEnumValue *value;
-  if (g_once_init_enter (&class))
-    g_once_init_leave(&class, (gpointer)g_type_class_ref (PANGO_TYPE_SCRIPT));
-
-  value = g_enum_get_value (class, script);
-  if (!value)
-    return string_from_script (PANGO_SCRIPT_INVALID_CODE);
-
-  return value->value_nick;
-}
-
 static void
 itemize_state_process_run (ItemizeState *state)
 {
@@ -1385,7 +1446,6 @@ itemize_state_process_run (ItemizeState *state)
     {
       gunichar wc = g_utf8_get_char (p);
       gboolean is_forced_break = (wc == '\t' || wc == LINE_SEPARATOR);
-      PangoEngineShape *shape_engine;
       PangoFont *font;
       GUnicodeType type;
 
@@ -1395,23 +1455,30 @@ itemize_state_process_run (ItemizeState *state)
        * characters if they don't, HarfBuzz will compatibility-decompose them
        * to ASCII space...
        * See bugs #355987 and #701652.
+       *
+       * We don't want to change fonts just for variation selectors.
+       * See bug #781123.
+       *
+       * Finally, don't change fonts for line or paragraph separators.
        */
       type = g_unichar_type (wc);
       if (G_UNLIKELY (type == G_UNICODE_CONTROL ||
                       type == G_UNICODE_FORMAT ||
                       type == G_UNICODE_SURROGATE ||
-                      type == G_UNICODE_SPACE_SEPARATOR))
+                      type == G_UNICODE_LINE_SEPARATOR ||
+                      type == G_UNICODE_PARAGRAPH_SEPARATOR ||
+                      (type == G_UNICODE_SPACE_SEPARATOR && wc != 0x1680u /* OGHAM SPACE MARK */) ||
+                      (wc >= 0xfe00u && wc <= 0xfe0fu) ||
+                      (wc >= 0xe0100u && wc <= 0xe01efu)))
         {
-	  shape_engine = NULL;
 	  font = NULL;
         }
       else
         {
-	  get_shaper_and_font (state, wc, &shape_engine, &font);
+	  get_font (state, wc, &font);
 	}
 
-      itemize_state_add_character (state,
-				   shape_engine, font,
+      itemize_state_add_character (state, font,
 				   is_forced_break || last_was_forced_break,
 				   p);
 
@@ -1420,32 +1487,30 @@ itemize_state_process_run (ItemizeState *state)
 
   /* Finish the final item from the current segment */
   state->item->length = (p - state->text) - state->item->offset;
-  if (!state->item->analysis.shape_engine)
+  if (!state->item->analysis.font)
     {
-      PangoEngineShape *shape_engine;
       PangoFont *font;
 
-      if (G_UNLIKELY (!get_shaper_and_font (state, ' ', &shape_engine, &font)))
-	{
-	  /* If no shaper was found, warn only once per fontmap/script pair */
-	  PangoFontMap *fontmap = state->context->font_map;
-	  const char *script_name = string_from_script (get_script (state));
+      if (G_UNLIKELY (!get_font (state, ' ', &font)))
+        {
+          /* If no font was found, warn once per fontmap/script pair */
+          PangoFontMap *fontmap = state->context->font_map;
+          char *script_tag = g_strdup_printf ("g-unicode-script-%d", state->script);
 
-	  if (!g_object_get_data (G_OBJECT (fontmap), script_name))
-	    {
-	      g_warning ("failed to choose a font, expect ugly output. engine-type='%s', script='%s'",
-			 pango_font_map_get_shape_engine_type (fontmap),
-			 script_name);
+          if (!g_object_get_data (G_OBJECT (fontmap), script_tag))
+            {
+              g_warning ("failed to choose a font, expect ugly output. script='%d'",
+                         state->script);
 
-	      g_object_set_data_full (G_OBJECT (fontmap), script_name,
-				      GINT_TO_POINTER (1), NULL);
-	    }
+              g_object_set_data_full (G_OBJECT (fontmap), script_tag,
+                                      GINT_TO_POINTER (1), NULL);
+            }
 
-	  shape_engine = _pango_get_fallback_shaper ();
-	  font = NULL;
-	}
+          g_free (script_tag);
 
-      itemize_state_fill_shaper (state, shape_engine, font);
+          font = NULL;
+        }
+      itemize_state_fill_font (state, font);
     }
   state->item = NULL;
 }
@@ -1458,6 +1523,9 @@ itemize_state_finish (ItemizeState *state)
     pango_attr_iterator_destroy (state->attr_iter);
   _pango_script_iter_fini (&state->script_iter);
   pango_font_description_free (state->font_desc);
+  pango_font_description_free (state->emoji_font_desc);
+  width_iter_fini (&state->width_iter);
+  _pango_emoji_iter_fini (&state->emoji_iter);
 
   if (state->current_fonts)
     g_object_unref (state->current_fonts);
@@ -1504,7 +1572,7 @@ pango_itemize_with_base_dir (PangoContext      *context,
   g_return_val_if_fail (length >= 0, NULL);
   g_return_val_if_fail (length == 0 || text != NULL, NULL);
 
-  if (length == 0)
+  if (length == 0 || g_utf8_strlen (text, length) == 0)
     return NULL;
 
   itemize_state_init (&state, context, text, base_dir, start_index, length,
@@ -1547,7 +1615,7 @@ itemize_with_font (PangoContext               *context,
  * pango_itemize:
  * @context:   a structure holding information that affects
 	       the itemization process.
- * @text:      the text to itemize.
+ * @text:      the text to itemize. Must be valid UTF-8
  * @start_index: first byte in @text to process
  * @length:    the number of bytes (not characters) to process
  *             after @start_index.
@@ -1633,6 +1701,10 @@ update_metrics_from_items (PangoFontMetrics *metrics,
   GHashTable *fonts_seen = g_hash_table_new (NULL, NULL);
   PangoGlyphString *glyphs = pango_glyph_string_new ();
   GList *l;
+  glong text_width;
+
+  /* This should typically be called with a sample text string. */
+  g_return_if_fail (text_len > 0);
 
   metrics->approximate_char_width = 0;
 
@@ -1649,6 +1721,7 @@ update_metrics_from_items (PangoFontMetrics *metrics,
 	  /* metrics will already be initialized from the first font in the fontset */
 	  metrics->ascent = MAX (metrics->ascent, raw_metrics->ascent);
 	  metrics->descent = MAX (metrics->descent, raw_metrics->descent);
+	  metrics->height = MAX (metrics->height, raw_metrics->height);
 	  pango_font_metrics_unref (raw_metrics);
 	}
 
@@ -1661,7 +1734,9 @@ update_metrics_from_items (PangoFontMetrics *metrics,
   pango_glyph_string_free (glyphs);
   g_hash_table_destroy (fonts_seen);
 
-  metrics->approximate_char_width /= pango_utf8_strwidth (text);
+  text_width = pango_utf8_strwidth (text);
+  g_assert (text_width > 0);
+  metrics->approximate_char_width /= text_width;
 }
 
 /**
@@ -1794,3 +1869,46 @@ pango_context_get_serial (PangoContext *context)
   check_fontmap_changed (context);
   return context->serial;
 }
+
+/**
+ * pango_context_set_round_glyph_positions:
+ * @context: a #PangoContext
+ * @round_positions: whether to round glyph positions
+ *
+ * Sets whether font rendering with this context should
+ * round glyph positions and widths to integral positions,
+ * in device units.
+ *
+ * This is useful when the renderer can't handle subpixel
+ * positioning of glyphs.
+ *
+ * The default value is to round glyph positions, to remain
+ * compatible with previous Pango behavior.
+ *
+ * Since: 1.44
+ */
+void
+pango_context_set_round_glyph_positions (PangoContext *context,
+                                         gboolean      round_positions)
+{
+  if (context->round_glyph_positions != round_positions)
+    {
+      context->round_glyph_positions = round_positions;
+      context_changed (context);
+    }
+}
+
+/**
+ * pango_context_get_round_glyph_positions:
+ * @context: a #PangoContext
+ *
+ * Returns whether font rendering with this context should
+ * round glyph positions and widths.
+ *
+ * Since: 1.44
+ */
+gboolean
+pango_context_get_round_glyph_positions (PangoContext *context)
+{
+  return context->round_glyph_positions;
+}
