@@ -66,7 +66,7 @@ pango_language_get_private (PangoLanguage *language)
   if (!language)
     return NULL;
 
-  priv = (PangoLanguagePrivate *) ((char *)language - sizeof (PangoLanguagePrivate));
+  priv = (PangoLanguagePrivate *)(void *)((char *)language - sizeof (PangoLanguagePrivate));
 
   if (G_UNLIKELY (priv->magic != PANGO_LANGUAGE_PRIVATE_MAGIC))
     {
@@ -638,6 +638,10 @@ pango_language_get_sample_string (PangoLanguage *language)
  * The pango_language_includes_script() function uses this function
  * internally.
  *
+ * Note: while the return value is declared as PangoScript, the
+ * returned values are from the GUnicodeScript enumeration, which
+ * may have more values. Callers need to handle unknown values.
+ *
  * Return value: (array length=num_scripts) (nullable): An array of
  * #PangoScript values, with the number of entries in the array stored
  * in @num_scripts, or %NULL if Pango does not have any information
@@ -658,7 +662,7 @@ pango_language_get_scripts (PangoLanguage *language,
 						 script_for_lang,
 						 pango_script_for_lang);
 
-  if (!script_for_lang)
+  if (!script_for_lang || script_for_lang->scripts[0] == 0)
     {
       if (num_scripts)
 	*num_scripts = 0;
@@ -677,7 +681,7 @@ pango_language_get_scripts (PangoLanguage *language,
       *num_scripts = j;
     }
 
-  return script_for_lang->scripts;
+  return (const PangoScript *) script_for_lang->scripts;
 }
 
 /**
