@@ -23,6 +23,20 @@
 
 #include "pango-ot-private.h"
 
+static PangoOTBuffer *
+pango_ot_buffer_copy (PangoOTBuffer *src)
+{
+  PangoOTBuffer *dst = g_slice_new (PangoOTBuffer);
+
+  dst->buffer = hb_buffer_reference (src->buffer);
+
+  return dst;
+}
+
+G_DEFINE_BOXED_TYPE (PangoOTBuffer, pango_ot_buffer,
+                     pango_ot_buffer_copy,
+                     pango_ot_buffer_destroy)
+
 /**
  * pango_ot_buffer_new
  * @font: a #PangoFcFont
@@ -40,7 +54,6 @@ pango_ot_buffer_new (PangoFcFont *font)
   PangoOTBuffer *buffer = g_slice_new (PangoOTBuffer);
 
   buffer->buffer = hb_buffer_create ();
-  hb_buffer_set_unicode_funcs (buffer->buffer, hb_glib_get_unicode_funcs ());
 
   return buffer;
 }
