@@ -26,7 +26,7 @@
 #include "pango-impl-utils.h"
 #include "pango-layout-private.h"
 
-#define N_RENDER_PARTS 4
+#define N_RENDER_PARTS 5
 
 #define PANGO_IS_RENDERER_FAST(renderer) (renderer != NULL)
 #define IS_VALID_PART(part) ((guint)part < N_RENDER_PARTS)
@@ -46,6 +46,10 @@ struct _LineState
 
   gboolean strikethrough;
   PangoRectangle strikethrough_rect;
+  int strikethrough_glyphs;
+
+  PangoOverline  overline;
+  PangoRectangle overline_rect;
 
   int logical_rect_end;
 };
@@ -58,6 +62,7 @@ struct _PangoRendererPrivate
 
   PangoLayoutLine *line;
   LineState *line_state;
+  PangoOverline overline;
 };
 
 static void pango_renderer_finalize                     (GObject          *gobject);
@@ -216,15 +221,17 @@ draw_underline (PangoRenderer *renderer,
     case PANGO_UNDERLINE_NONE:
       break;
     case PANGO_UNDERLINE_DOUBLE:
+    case PANGO_UNDERLINE_DOUBLE_LINE:
       pango_renderer_draw_rectangle (renderer,
 				     PANGO_RENDER_PART_UNDERLINE,
 				     rect->x,
 				     rect->y + 2 * rect->height,
 				     rect->width,
 				     rect->height);
-      /* Fall through */
+      G_GNUC_FALLTHROUGH;
     case PANGO_UNDERLINE_SINGLE:
     case PANGO_UNDERLINE_LOW:
+    case PANGO_UNDERLINE_SINGLE_LINE:
       pango_renderer_draw_rectangle (renderer,
 				     PANGO_RENDER_PART_UNDERLINE,
 				     rect->x,
@@ -233,6 +240,7 @@ draw_underline (PangoRenderer *renderer,
 				     rect->height);
       break;
     case PANGO_UNDERLINE_ERROR:
+    case PANGO_UNDERLINE_ERROR_LINE:
       pango_renderer_draw_error_underline (renderer,
 					   rect->x,
 					   rect->y,
@@ -242,22 +250,47 @@ draw_underline (PangoRenderer *renderer,
     }
 }
 
+static void
+draw_overline (PangoRenderer *renderer,
+               LineState     *state)
+{
+  PangoRectangle *rect = &state->overline_rect;
+  PangoOverline overline = state->overline;
+
+  state->overline = PANGO_OVERLINE_NONE;
+
+  switch (overline)
+    {
+    case PANGO_OVERLINE_NONE:
+      break;
+    case PANGO_OVERLINE_SINGLE:
+      pango_renderer_draw_rectangle (renderer,
+				     PANGO_RENDER_PART_OVERLINE,
+				     rect->x,
+				     rect->y,
+				     rect->width,
+				     rect->height);
+      break;
+    }
+}
+
 static void
 draw_strikethrough (PangoRenderer *renderer,
 		    LineState     *state)
 {
   PangoRectangle *rect = &state->strikethrough_rect;
   gboolean strikethrough = state->strikethrough;
+  int num_glyphs = state->strikethrough_glyphs;
 
   state->strikethrough = FALSE;
 
   if (strikethrough)
     pango_renderer_draw_rectangle (renderer,
-				   PANGO_RENDER_PART_STRIKETHROUGH,
-				   rect->x,
-				   rect->y,
-				   rect->width,
-				   rect->height);
+                                   PANGO_RENDER_PART_STRIKETHROUGH,
+                                   rect->x,
+                                   rect->y / num_glyphs,
+                                   rect->width,
+                                   rect->height / num_glyphs);
 }
 
 static void
@@ -280,6 +313,18 @@ handle_line_state_change (PangoRenderer  *renderer,
       rect->width = 0;
     }
 
+  if (part == PANGO_RENDER_PART_OVERLINE &&
+      state->overline != PANGO_OVERLINE_NONE)
+    {
+      PangoRectangle *rect = &state->overline_rect;
+
+      rect->width = state->logical_rect_end - rect->x;
+      draw_overline (renderer, state);
+      state->overline = renderer->priv->overline;
+      rect->x = state->logical_rect_end;
+      rect->width = 0;
+    }
+
   if (part == PANGO_RENDER_PART_STRIKETHROUGH &&
       state->strikethrough)
     {
@@ -326,6 +371,18 @@ add_underline (PangoRenderer    *renderer,
     case PANGO_UNDERLINE_LOW:
       new_rect.y += ink_rect->y + ink_rect->height + underline_thickness;
       break;
+    case PANGO_UNDERLINE_SINGLE_LINE:
+    case PANGO_UNDERLINE_DOUBLE_LINE:
+    case PANGO_UNDERLINE_ERROR_LINE:
+      new_rect.y -= underline_position;
+      if (state->underline == renderer->underline)
+        {
+          new_rect.y = MAX (current_rect->y, new_rect.y);
+          new_rect.height = MAX (current_rect->height, new_rect.height);
+          current_rect->y = new_rect.y;
+          current_rect->height = new_rect.height;
+        }
+      break;
     }
 
   if (renderer->underline == state->underline &&
@@ -343,6 +400,57 @@ add_underline (PangoRenderer    *renderer,
     }
 }
 
+static void
+add_overline (PangoRenderer    *renderer,
+              LineState        *state,
+              PangoFontMetrics *metrics,
+              int               base_x,
+              int               base_y,
+              PangoRectangle   *ink_rect,
+              PangoRectangle   *logical_rect)
+{
+  PangoRectangle *current_rect = &state->overline_rect;
+  PangoRectangle new_rect;
+  int underline_thickness = pango_font_metrics_get_underline_thickness (metrics);
+  int ascent = pango_font_metrics_get_ascent (metrics);
+
+  new_rect.x = base_x + logical_rect->x;
+  new_rect.width = logical_rect->width;
+  new_rect.height = underline_thickness;
+  new_rect.y = base_y;
+
+  switch (renderer->priv->overline)
+    {
+    case PANGO_OVERLINE_NONE:
+      g_assert_not_reached ();
+      break;
+    case PANGO_OVERLINE_SINGLE:
+      new_rect.y -= ascent;
+      if (state->overline == renderer->priv->overline)
+        {
+          new_rect.y = MIN (current_rect->y, new_rect.y);
+          new_rect.height = MAX (current_rect->height, new_rect.height);
+          current_rect->y = new_rect.y;
+          current_rect->height = new_rect.height;
+        }
+      break;
+    }
+
+  if (renderer->priv->overline == state->overline &&
+      new_rect.y == current_rect->y &&
+      new_rect.height == current_rect->height)
+    {
+      current_rect->width = new_rect.x + new_rect.width - current_rect->x;
+    }
+  else
+    {
+      draw_overline (renderer, state);
+
+      *current_rect = new_rect;
+      state->overline = renderer->priv->overline;
+    }
+}
+
 static void
 add_strikethrough (PangoRenderer    *renderer,
 		   LineState        *state,
@@ -350,7 +458,8 @@ add_strikethrough (PangoRenderer    *renderer,
 		   int               base_x,
 		   int               base_y,
 		   PangoRectangle   *ink_rect G_GNUC_UNUSED,
-		   PangoRectangle   *logical_rect)
+		   PangoRectangle   *logical_rect,
+                   int               num_glyphs)
 {
   PangoRectangle *current_rect = &state->strikethrough_rect;
   PangoRectangle new_rect;
@@ -360,21 +469,22 @@ add_strikethrough (PangoRenderer    *renderer,
 
   new_rect.x = base_x + logical_rect->x;
   new_rect.width = logical_rect->width;
-  new_rect.y = base_y - strikethrough_position;
-  new_rect.height = strikethrough_thickness;
+  new_rect.y = (base_y - strikethrough_position) * num_glyphs;
+  new_rect.height = strikethrough_thickness * num_glyphs;
 
-  if (state->strikethrough &&
-      new_rect.y == current_rect->y &&
-      new_rect.height == current_rect->height)
+  if (state->strikethrough)
     {
       current_rect->width = new_rect.x + new_rect.width - current_rect->x;
+      current_rect->y += new_rect.y;
+      current_rect->height += new_rect.height;
+      state->strikethrough_glyphs += num_glyphs;
     }
   else
     {
       draw_strikethrough (renderer, state);
-
       *current_rect = new_rect;
       state->strikethrough = TRUE;
+      state->strikethrough_glyphs = num_glyphs;
     }
 }
 
@@ -482,6 +592,7 @@ pango_renderer_draw_layout_line (PangoRenderer    *renderer,
   renderer->priv->line_state = &state;
 
   state.underline = PANGO_UNDERLINE_NONE;
+  state.overline = PANGO_OVERLINE_NONE;
   state.strikethrough = FALSE;
 
   text = G_LIKELY (line->layout) ? pango_layout_get_text (line->layout) : NULL;
@@ -516,6 +627,7 @@ pango_renderer_draw_layout_line (PangoRenderer    *renderer,
       else
 	{
 	  if (renderer->underline != PANGO_UNDERLINE_NONE ||
+              renderer->priv->overline != PANGO_OVERLINE_NONE ||
 	      renderer->strikethrough)
 	    {
 	      ink = &ink_rect;
@@ -573,6 +685,7 @@ pango_renderer_draw_layout_line (PangoRenderer    *renderer,
 	}
 
       if (renderer->underline != PANGO_UNDERLINE_NONE ||
+          renderer->priv->overline != PANGO_OVERLINE_NONE ||
 	  renderer->strikethrough)
 	{
 	  metrics = pango_font_get_metrics (run->item->analysis.font,
@@ -583,10 +696,15 @@ pango_renderer_draw_layout_line (PangoRenderer    *renderer,
 			   x + x_off, y - rise,
 			   ink, logical);
 
+	  if (renderer->priv->overline != PANGO_OVERLINE_NONE)
+	    add_overline (renderer, &state,metrics,
+			   x + x_off, y - rise,
+			   ink, logical);
+
 	  if (renderer->strikethrough)
 	    add_strikethrough (renderer, &state, metrics,
 			       x + x_off, y - rise,
-			       ink, logical);
+			       ink, logical, run->glyphs->num_glyphs);
 
 	  pango_font_metrics_unref (metrics);
 	}
@@ -595,6 +713,10 @@ pango_renderer_draw_layout_line (PangoRenderer    *renderer,
 	  state.underline != PANGO_UNDERLINE_NONE)
 	draw_underline (renderer, &state);
 
+      if (renderer->priv->overline == PANGO_OVERLINE_NONE &&
+	  state.overline != PANGO_OVERLINE_NONE)
+	draw_overline (renderer, &state);
+
       if (!renderer->strikethrough && state.strikethrough)
 	draw_strikethrough (renderer, &state);
 
@@ -604,6 +726,7 @@ pango_renderer_draw_layout_line (PangoRenderer    *renderer,
   /* Finish off any remaining underlines
    */
   draw_underline (renderer, &state);
+  draw_overline (renderer, &state);
   draw_strikethrough (renderer, &state);
 
   renderer->priv->line_state = NULL;
@@ -974,15 +1097,22 @@ pango_renderer_default_draw_error_underline (PangoRenderer *renderer,
 					     int            width,
 					     int            height)
 {
-  int square = height / HEIGHT_SQUARES;
-  int unit_width = (HEIGHT_SQUARES - 1) * square;
-  int width_units = (width + unit_width / 2) / unit_width;
+  int square;
+  int unit_width;
+  int width_units;
   const PangoMatrix identity = PANGO_MATRIX_INIT;
   const PangoMatrix *matrix;
   double dx, dx0, dy0;
   PangoMatrix total;
   int i;
 
+  if (width <= 0 || height <= 0)
+    return;
+
+  square = height / HEIGHT_SQUARES;
+  unit_width = (HEIGHT_SQUARES - 1) * square;
+  width_units = (width + unit_width / 2) / unit_width;
+
   x += (width - width_units * unit_width) / 2;
 
   if (renderer->matrix)
@@ -1313,12 +1443,14 @@ pango_renderer_default_prepare_run (PangoRenderer  *renderer,
   PangoColor *fg_color = NULL;
   PangoColor *bg_color = NULL;
   PangoColor *underline_color = NULL;
+  PangoColor *overline_color = NULL;
   PangoColor *strikethrough_color = NULL;
   guint16 fg_alpha = 0;
   guint16 bg_alpha = 0;
   GSList *l;
 
   renderer->underline = PANGO_UNDERLINE_NONE;
+  renderer->priv->overline = PANGO_OVERLINE_NONE;
   renderer->strikethrough = FALSE;
 
   for (l = run->item->analysis.extra_attrs; l; l = l->next)
@@ -1331,6 +1463,10 @@ pango_renderer_default_prepare_run (PangoRenderer  *renderer,
 	  renderer->underline = ((PangoAttrInt *)attr)->value;
 	  break;
 
+	case PANGO_ATTR_OVERLINE:
+	  renderer->priv->overline = ((PangoAttrInt *)attr)->value;
+	  break;
+
 	case PANGO_ATTR_STRIKETHROUGH:
 	  renderer->strikethrough = ((PangoAttrInt *)attr)->value;
 	  break;
@@ -1347,6 +1483,10 @@ pango_renderer_default_prepare_run (PangoRenderer  *renderer,
 	  underline_color = &((PangoAttrColor *)attr)->color;
 	  break;
 
+	case PANGO_ATTR_OVERLINE_COLOR:
+	  overline_color = &((PangoAttrColor *)attr)->color;
+	  break;
+
 	case PANGO_ATTR_STRIKETHROUGH_COLOR:
 	  strikethrough_color = &((PangoAttrColor *)attr)->color;
 	  break;
@@ -1367,6 +1507,9 @@ pango_renderer_default_prepare_run (PangoRenderer  *renderer,
   if (!underline_color)
     underline_color = fg_color;
 
+  if (!overline_color)
+    overline_color = fg_color;
+
   if (!strikethrough_color)
     strikethrough_color = fg_color;
 
@@ -1374,11 +1517,13 @@ pango_renderer_default_prepare_run (PangoRenderer  *renderer,
   pango_renderer_set_color (renderer, PANGO_RENDER_PART_BACKGROUND, bg_color);
   pango_renderer_set_color (renderer, PANGO_RENDER_PART_UNDERLINE, underline_color);
   pango_renderer_set_color (renderer, PANGO_RENDER_PART_STRIKETHROUGH, strikethrough_color);
+  pango_renderer_set_color (renderer, PANGO_RENDER_PART_OVERLINE, overline_color);
 
   pango_renderer_set_alpha (renderer, PANGO_RENDER_PART_FOREGROUND, fg_alpha);
   pango_renderer_set_alpha (renderer, PANGO_RENDER_PART_BACKGROUND, bg_alpha);
   pango_renderer_set_alpha (renderer, PANGO_RENDER_PART_UNDERLINE, fg_alpha);
   pango_renderer_set_alpha (renderer, PANGO_RENDER_PART_STRIKETHROUGH, fg_alpha);
+  pango_renderer_set_alpha (renderer, PANGO_RENDER_PART_OVERLINE, fg_alpha);
 }
 
 /**
