@@ -32,6 +32,7 @@ typedef struct _PangoCairoContextInfo PangoCairoContextInfo;
 struct _PangoCairoContextInfo
 {
   double dpi;
+  gboolean set_options_explicit;
 
   cairo_font_options_t *set_options;
   cairo_font_options_t *surface_options;
@@ -108,6 +109,12 @@ _pango_cairo_update_context (cairo_t      *cr,
   if (!info->surface_options)
     info->surface_options = cairo_font_options_create ();
   cairo_surface_get_font_options (target, info->surface_options);
+  if (!info->set_options_explicit)
+  {
+    if (!info->set_options)
+      info->set_options = cairo_font_options_create ();
+    cairo_get_font_options (cr, info->set_options);
+  }
 
   old_merged_options = info->merged_options;
   info->merged_options = NULL;
@@ -235,7 +242,15 @@ pango_cairo_context_set_font_options (PangoContext               *context,
 
   g_return_if_fail (PANGO_IS_CONTEXT (context));
 
-  info  = get_context_info (context, TRUE);
+  info = get_context_info (context, TRUE);
+
+  if (!info->set_options && !options)
+    return;
+
+  if (info->set_options &&
+      options &&
+      cairo_font_options_equal (info->set_options, options))
+    return;
 
   if (info->set_options || options)
     pango_context_changed (context);
@@ -244,9 +259,15 @@ pango_cairo_context_set_font_options (PangoContext               *context,
     cairo_font_options_destroy (info->set_options);
 
   if (options)
+  {
     info->set_options = cairo_font_options_copy (options);
+    info->set_options_explicit = TRUE;
+  }
   else
+  {
     info->set_options = NULL;
+    info->set_options_explicit = FALSE;
+  }
 
   if (info->merged_options)
     {
@@ -361,7 +382,7 @@ pango_cairo_context_set_shape_renderer (PangoContext                *context,
  * attributes of type %PANGO_ATTR_SHAPE as set by
  * pango_cairo_context_set_shape_renderer(), if any.
  *
- * Return value: (nullable): the shape rendering callback previously
+ * Return value: (transfer none) (nullable): the shape rendering callback previously
  *   set on the context, or %NULL if no shape rendering callback have
  *   been set.
  *
