@@ -103,7 +103,7 @@ pango_cairo_fc_font_map_fontset_key_substitute (PangoFcFontMap    *fcfontmap G_G
 						PangoFcFontsetKey *fontkey,
 						FcPattern         *pattern)
 {
-  FcConfigSubstitute (NULL, pattern, FcMatchPattern);
+  FcConfigSubstitute (pango_fc_font_map_get_config (fcfontmap), pattern, FcMatchPattern);
 
   if (fontkey)
     cairo_ft_font_options_substitute (pango_fc_fontset_key_get_context_key (fontkey),
