@@ -331,7 +331,6 @@ get_glyph_extents_raw (PangoXftFont     *xfont,
       extents = g_slice_new (Extents);
 
       pango_fc_font_get_raw_extents (PANGO_FC_FONT (xfont),
-				     FT_LOAD_NO_BITMAP | FT_LOAD_NO_HINTING,
 				     glyph,
 				     &extents->ink_rect,
 				     &extents->logical_rect);
@@ -500,13 +499,13 @@ pango_xft_font_real_shutdown (PangoFcFont *fcfont)
 }
 
 /**
- * pango_xft_font_get_font:
+ * pango_xft_font_get_font: (skip)
  * @font: (nullable): a #PangoFont.
  *
- * Returns the XftFont of a font.
+ * Returns the `XftFont` of a font.
  *
- * Return value: (nullable): the XftFont associated to @font, or %NULL
- * if @font is %NULL.
+ * Return value: (nullable): the `XftFont` associated to @font,
+ *   or %NULL if @font is %NULL.
  **/
 XftFont *
 pango_xft_font_get_font (PangoFont *font)
@@ -518,12 +517,12 @@ pango_xft_font_get_font (PangoFont *font)
 }
 
 /**
- * pango_xft_font_get_display:
- * @font: a #PangoFont.
+ * pango_xft_font_get_display: (skip)
+ * @font: (type PangoXftFont): a #PangoFont.
  *
- * Returns the X display of the XftFont of a font.
+ * Returns the X display of the `XftFont` of a font.
  *
- * Return value: the X display of the XftFont associated to @font.
+ * Return value: (transfer none): the X display of the XftFont associated to @font.
  **/
 Display *
 pango_xft_font_get_display (PangoFont *font)
@@ -541,7 +540,7 @@ pango_xft_font_get_display (PangoFont *font)
 
 /**
  * pango_xft_font_get_unknown_glyph:
- * @font: a #PangoFont.
+ * @font: (type PangoXftFont): a #PangoFont.
  * @wc: the Unicode character for which a glyph is needed.
  *
  * Returns the index of a glyph suitable for drawing @wc as an
@@ -561,16 +560,16 @@ pango_xft_font_get_unknown_glyph (PangoFont *font,
 }
 
 /**
- * pango_xft_font_lock_face:
- * @font: a #PangoFont.
+ * pango_xft_font_lock_face: (skip)
+ * @font: (type PangoXftFont): a #PangoFont.
  *
- * Gets the FreeType <type>FT_Face</type> associated with a font,
- * This face will be kept around until you call
- * pango_xft_font_unlock_face().
+ * Gets the FreeType `FT_Face` associated with a font.
+ *
+ * This face will be kept around until you call pango_xft_font_unlock_face().
  *
  * Use pango_fc_font_lock_face() instead.
  *
- * Return value: the FreeType <type>FT_Face</type> associated with @font.
+ * Return value: the FreeType `FT_Face` associated with @font.
  *
  * Since: 1.2
  **/
@@ -583,8 +582,8 @@ pango_xft_font_lock_face (PangoFont *font)
 }
 
 /**
- * pango_xft_font_unlock_face:
- * @font: a #PangoFont.
+ * pango_xft_font_unlock_face: (skip)
+ * @font: (type PangoXftFont): a #PangoFont.
  *
  * Releases a font previously obtained with
  * pango_xft_font_lock_face().
@@ -603,7 +602,7 @@ pango_xft_font_unlock_face (PangoFont *font)
 
 /**
  * pango_xft_font_get_glyph:
- * @font: a #PangoFont for the Xft backend
+ * @font: (type PangoXftFont): a #PangoFont for the Xft backend
  * @wc: Unicode codepoint to look up
  *
  * Gets the glyph index for a given Unicode character
@@ -628,7 +627,7 @@ pango_xft_font_get_glyph (PangoFont *font,
 
 /**
  * pango_xft_font_has_char:
- * @font: a #PangoFont for the Xft backend
+ * @font: (type PangoXftFont): a #PangoFont for the Xft backend
  * @wc: Unicode codepoint to look up
  *
  * Determines whether @font has a glyph for the codepoint @wc.
