@@ -34,7 +34,7 @@
 
 static PangoContext *context;
 
-static void
+static gboolean
 test_file (const gchar *filename, GString *string)
 {
   gchar *contents;
@@ -51,12 +51,8 @@ test_file (const gchar *filename, GString *string)
   PangoAttrList *attributes;
   PangoLayout *layout;
 
-  if (!g_file_get_contents (filename, &contents, &length, &error))
-    {
-      fprintf (stderr, "%s\n", error->message);
-      g_error_free (error);
-      return;
-    }
+  g_file_get_contents (filename, &contents, &length, &error);
+  g_assert_no_error (error);
 
   test = contents;
 
@@ -66,18 +62,29 @@ test_file (const gchar *filename, GString *string)
 
   length = strlen (test);
   len = g_utf8_strlen (test, -1) + 1;
-  attrs = g_new (PangoLogAttr, len);
 
-  if (!pango_parse_markup (test, -1, 0, &attributes, &text, NULL, &error))
-    {
-      fprintf (stderr, "%s\n", error->message);
-      g_error_free (error);
-      return;
-    }
+  pango_parse_markup (test, -1, 0, &attributes, &text, NULL, &error);
+  g_assert_no_error (error);
 
   layout = pango_layout_new (context);
   pango_layout_set_text (layout, text, length);
   pango_layout_set_attributes (layout, attributes);
+
+  if (pango_layout_get_unknown_glyphs_count (layout) > 0)
+    {
+#if 0
+      // See https://github.com/mesonbuild/meson/issues/7515
+      char *msg = g_strdup_printf ("Missing glyphs - skipping %s. Maybe fonts are missing?", filename);
+      g_test_skip (msg);
+      g_free (msg);
+#endif
+      g_free (contents);
+      g_object_unref (layout);
+      pango_attr_list_unref (attributes);
+      g_free (text);
+      return FALSE;
+    }
+
   pango_layout_get_log_attrs (layout, &attrs, &len);
 
   s1 = g_string_new ("Breaks: ");
@@ -221,7 +228,10 @@ test_file (const gchar *filename, GString *string)
   g_object_unref (layout);
   g_free (attrs);
   g_free (contents);
+  g_free (text);
   pango_attr_list_unref (attributes);
+
+  return TRUE;
 }
 
 static gchar *
@@ -249,22 +259,54 @@ test_break (gconstpointer d)
   GString *dump;
   gchar *diff;
 
-  expected_file = get_expected_filename (filename);
+  char *old_locale = g_strdup (setlocale (LC_ALL, NULL));
+  setlocale (LC_ALL, "en_US.utf8");
+  if (strstr (setlocale (LC_ALL, NULL), "en_US") == NULL)
+    {
+#if 0
+      // See https://github.com/mesonbuild/meson/issues/7515
+      char *msg = g_strdup_printf ("Locale en_US.UTF-8 not available, skipping break %s", filename);
+      g_test_skip (msg);
+      g_free (msg);
+#endif
+      g_free (old_locale);
+      return;
+    }
 
   dump = g_string_sized_new (0);
 
-  test_file (filename, dump);
+  if (!test_file (filename, dump))
+    {
+      g_free (old_locale);
+      g_string_free (dump, TRUE);
+      return;
+    }
+
+  expected_file = get_expected_filename (filename);
 
   diff = diff_with_file (expected_file, dump->str, dump->len, &error);
   g_assert_no_error (error);
 
+  setlocale (LC_ALL, old_locale);
+  g_free (old_locale);
+
   if (diff && diff[0])
     {
-      g_printerr ("Contents don't match expected contents:\n%s", diff);
+      char **lines = g_strsplit (diff, "\n", -1);
+      const char *line;
+      int i = 0;
+
+      g_test_message ("Contents don't match expected contents");
+
+      for (line = lines[0]; line != NULL; line = lines[++i])
+        g_test_message ("%s", line);
+
       g_test_fail ();
-      g_free (diff);
+
+      g_strfreev (lines);
     }
 
+  g_free (diff);
   g_string_free (dump, TRUE);
   g_free (expected_file);
 }
@@ -277,21 +319,35 @@ main (int argc, char *argv[])
   const gchar *name;
   gchar *path;
 
-  g_setenv ("LC_ALL", "en_US.UTF-8", TRUE);
-  setlocale (LC_ALL, "");
-
   g_test_init (&argc, &argv, NULL);
 
+  setlocale (LC_ALL, "");
+
   context = pango_font_map_create_context (pango_cairo_font_map_get_default ());
 
   /* allow to easily generate expected output for new test cases */
   if (argc > 1)
     {
-      GString *string;
+      if (strcmp (argv[1], "--help") == 0)
+        {
+          g_print ("test-break uses the following symbols for log attrs\n\n");
+          g_print ("Breaks:                 Words:\n"
+                   " L - mandatory break     b - word boundary\n"
+                   " l - line break          s - word start\n"
+                   " c - char break          e - word end\n"
+                   "\n"
+                   "Whitespace:             Sentences:\n"
+                   " x - expandable space    s - sentence start\n"
+                   " w - whitespace          e - sentence end\n");
+        }
+      else
+        {
+          GString *string;
 
-      string = g_string_sized_new (0);
-      test_file (argv[1], string);
-      printf ("%s", string->str);
+          string = g_string_sized_new (0);
+          test_file (argv[1], string);
+          g_print ("%s", string->str);
+        }
 
       return 0;
     }
@@ -305,6 +361,12 @@ main (int argc, char *argv[])
       if (!strstr (name, "break"))
         continue;
 
+#ifndef HAVE_LIBTHAI
+      /* four.break involves Thai, so only test it when we have libthai */
+      if (strstr (name, "four.break"))
+        continue;
+#endif
+
       path = g_strdup_printf ("/break/%s", name);
       g_test_add_data_func_full (path, g_test_build_filename (G_TEST_DIST, "breaks", name, NULL),
                                  test_break, g_free);
