@@ -110,7 +110,7 @@ test_file (const gchar *filename, GString *string)
   gchar *contents;
   gsize  length;
   GError *error = NULL;
- GString *s1, *s2, *s3, *s4, *s5, *s6;
+  GString *s1, *s2, *s3, *s4, *s5, *s6;
   char *test;
   char *text;
   PangoAttrList *attrs;
@@ -118,12 +118,8 @@ test_file (const gchar *filename, GString *string)
   GList *items, *l;
   const char *sep = "";
 
-  if (!g_file_get_contents (filename, &contents, &length, &error))
-    {
-      fprintf (stderr, "%s\n", error->message);
-      g_error_free (error);
-      return;
-    }
+  g_file_get_contents (filename, &contents, &length, &error);
+  g_assert_no_error (error);
 
   test = contents;
 
@@ -131,13 +127,8 @@ test_file (const gchar *filename, GString *string)
   while (test[0] == '#')
     test = strchr (test, '\n') + 1;
 
-
-  if (!pango_parse_markup (test, -1, 0, &attrs, &text, NULL, &error))
-    {
-      fprintf (stderr, "%s\n", error->message);
-      g_error_free (error);
-      return;
-    }
+  pango_parse_markup (test, -1, 0, &attrs, &text, NULL, &error);
+  g_assert_no_error (error);
 
   s1 = g_string_new ("Items:  ");
   s2 = g_string_new ("Font:   ");
@@ -246,6 +237,21 @@ test_itemize (gconstpointer d)
   GString *dump;
   gchar *diff;
 
+  const char *old_locale = setlocale (LC_ALL, NULL);
+  setlocale (LC_ALL, "en_US.utf8");
+  if (strstr (setlocale (LC_ALL, NULL), "en_US") == NULL)
+    {
+#if 0
+      // See https://github.com/mesonbuild/meson/issues/7515
+      char *msg = g_strdup_printf ("Locale en_US.UTF-8 not available, skipping itemization %s", filename);
+      g_test_skip (msg);
+      g_free (msg);
+#endif
+      return;
+    }
+
+  context = pango_font_map_create_context (pango_cairo_font_map_get_default ());
+
   expected_file = get_expected_filename (filename);
 
   dump = g_string_sized_new (0);
@@ -255,13 +261,25 @@ test_itemize (gconstpointer d)
   diff = diff_with_file (expected_file, dump->str, dump->len, &error);
   g_assert_no_error (error);
 
+  setlocale (LC_ALL, old_locale);
+
   if (diff && diff[0])
     {
-      g_printerr ("Contents don't match expected contents:\n%s", diff);
+      char **lines = g_strsplit (diff, "\n", -1);
+      const char *line;
+      int i = 0;
+
+      g_test_message ("Contents don't match expected contents");
+
+      for (line = lines[0]; line != NULL; line = lines[++i])
+        g_test_message ("%s", line);
+
       g_test_fail ();
-      g_free (diff);
+
+      g_strfreev (lines);
     }
 
+  g_free (diff);
   g_string_free (dump, TRUE);
   g_free (expected_file);
 }
@@ -274,13 +292,8 @@ main (int argc, char *argv[])
   const gchar *name;
   gchar *path;
 
-  g_setenv ("LC_ALL", "en_US.UTF-8", TRUE);
-  setlocale (LC_ALL, "");
-
   g_test_init (&argc, &argv, NULL);
 
-  context = pango_font_map_create_context (pango_cairo_font_map_get_default ());
-
   /* allow to easily generate expected output for new test cases */
   if (argc > 1)
     {
